/*
 * Copyright (C) 2021- TeraTerm Project
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* String.h ɍ쐬 */

#ifndef _YCL_WSTRING_H_
#define _YCL_WSTRING_H_

#pragma once

#include <YCL/common.h>

#include <string.h>
#include <wchar.h>

namespace yebisuya {

// ̊ǗEsNXB
class WString {
private:
	// i[obt@B
	// ̑Oɂ͎QƃJE^A
	// j̍ۂɂ͂ύXB
	const wchar_t* string;

	// utilities
	// i[obt@쐬B
	// ƎQƃJE^̗̈̕mۂB
	// QƃJE^0ɂȂĂB
	// :
	//	length ̒B
	// Ԓl:
	//	쐬obt@̕񕔂̃AhXB
	static wchar_t* createBuffer(size_t length) {
		size_t* count = (size_t*) new unsigned char[sizeof (size_t) + sizeof (wchar_t) * (length + 1)];
		*count = 0;
		return (wchar_t*) (count + 1);
	}
	// i[obt@쐬B
	// :
	//	source i[镶B
	// Ԓl:
	//	쐬obt@̕񕔂̃AhXB
	static const wchar_t* create(const wchar_t* source) {
		return source != NULL ? create(source, wcslen(source)) : NULL;
	}
	// i[obt@쐬B
	// :
	//	source i[镶B
	//	length ̒B
	// Ԓl:
	//	쐬obt@̕񕔂̃AhXB
	static const wchar_t* create(const wchar_t* source, size_t length) {
		if (source != NULL) {
			wchar_t* buffer = createBuffer(length);
			wmemcpy(buffer, source, length);
			buffer[length] = '\0';
			return buffer;
		}
		return NULL;
	}
	// ̕Ai[obt@쐬B
	// :
	//	str1 A镶(O)B
	//	str2 A镶()B
	// Ԓl:
	//	쐬obt@̕񕔂̃AhXB
	static const wchar_t* concat(const wchar_t* str1, const wchar_t* str2) {
		size_t len1 = wcslen(str1);
		size_t len2 = wcslen(str2);
		wchar_t* buffer = createBuffer(len1 + len2);
		wmemcpy(buffer, str1, len1);
		wmemcpy(buffer + len1, str2, len2);
		buffer[len1 + len2] = '\0';
		return buffer;
	}
	// private methods
	// QƃJE^炵A0ɂȂobt@jB
	void release() {
		if (string != NULL) {
			size_t* count = (size_t*) string - 1;
			if (--*count == 0)
				delete[] (unsigned char*) count;
		}
	}
	// QƃJE^𑝂₷B
	void add() {
		if (string != NULL) {
			size_t* count = (size_t*) string - 1;
			++*count;
		}
	}
	// ʂ̃obt@ƒuB
	// ̃obt@̎QƃJE^炵A
	// Vobt@̎QƃJE^𑝂₷B
	// :
	//	source uVobt@B
	void set(const wchar_t* source) {
		if (string != source) {
			release();
			string = source;
			add();
		}
	}
public:
	// constructor
	// ftHgRXgN^B
	// NULLĂ̂ŁÂ܂܂ŕ񑀍삷ƃANZXᔽɂȂ̂ŒӁB
	WString():string(NULL) {
	}
	// ̕w肷RXgN^B
	// :
	//	source ̕B
	WString(const wchar_t* source):string(NULL) {
		set(create(source));
	}
	// ̕𒷂tŎw肷RXgN^B
	// :
	//	source ̕B
	//	length ̒B
	WString(const wchar_t* source, size_t length):string(NULL) {
		set(create(source, length));
	}
	// Rs[RXgN^B
	// :
	//	source ̕B
	WString(const WString& source):string(NULL) {
		set(source.string);
	}
	// ̕ARXgN^B
	// :
	//	str1 OɂȂ镶B
	//	str2 ɂȂ镶B
	WString(const wchar_t* str1, const wchar_t* str2):string(NULL) {
		set(concat(str1, str2));
	}
	// ̕ARXgN^B
	// :
	//	str1 OɂȂ镶B
	//	str2 ɂȂ镶B
	WString(const WString& str1, const wchar_t* str2):string(NULL) {
		set(*str2 != '\0' ? concat(str1.string, str2) : str1.string);
	}
	// destructor
	// fXgN^B
	// h邱Ƃ͍lĂȂ̂ŉz֐ɂ͂ȂB
	~WString() {
		release();
	}
	// public methods
	// ̌̕Ɏw̕AB
	// :
	//	source A镶B
	// Ԓl:
	//	AꂽB
	WString concat(const wchar_t* source)const {
		return WString(*this, source);
	}
	// Ƃ̔rsB
	// NULLƂrłB
	// :
	//	str r镶B
	// Ԓl:
	//	0Astr̕傫ΕAΐB
	int compareTo(const wchar_t* str)const {
		if (str == NULL)
			return string == NULL ? 0 : 1;
		else if (string == NULL)
			return -1;
		return wcscmp(string, str);
	}
	// Ƃ̔r啶̋ʂȂōsB
	// NULLƂrłB
	// :
	//	str r镶B
	// Ԓl:
	//	0Astr̕傫ΕAΐB
	int compareToIgnoreCase(const wchar_t* str)const {
		if (str == NULL)
			return string == NULL ? 0 : 1;
		else if (string == NULL)
			return -1;
		return _wcsicmp(string, str);
	}
	// Ƃ̔rsB
	// NULLƂrłB
	// :
	//	str r镶B
	// Ԓl:
	//	ΐ^B
	bool equals(const wchar_t* str)const {
		return compareTo(str) == 0;
	}
	// Ƃ̔r啶̋ʂȂōsB
	// NULLƂrłB
	// :
	//	str r镶B
	// Ԓl:
	//	ΐ^B
	bool equalsIgnoreCase(const wchar_t* str)const {
		return compareToIgnoreCase(str) == 0;
	}
	// w肳ꂽŎn܂Ă邩ǂ𔻒肷B
	// :
	//	str r镶B
	// Ԓl:
	//	w肳ꂽŎn܂Ăΐ^B
	bool startsWith(const wchar_t* str)const {
		return startsWith(str, 0);
	}
	// ẅʒuw肳ꂽŎn܂Ă邩ǂ𔻒肷B
	// :
	//	str r镶B
	// Ԓl:
	//	w肳ꂽŎn܂Ăΐ^B
	bool startsWith(const wchar_t* str, int offset)const {
		return wcsncmp(string, str, wcslen(str)) == 0;
	}
	// w肳ꂽŏIĂ邩ǂ𔻒肷B
	// :
	//	str r镶B
	// Ԓl:
	//	w肳ꂽŏIĂΐ^B
	//
	bool endsWith(const wchar_t* str)const {
		size_t str_length = wcslen(str);
		size_t string_length = length();
		if (string_length < str_length)
			return false;
		return wcscmp(string + string_length - str_length, str) == 0;
	}
	// w̕ǂ̈ʒuɂ邩TB
	// :
	//	chr TB
	// Ԓl:
	//	̌CfbNXBȂ-1B
	int indexOf(char chr)const {
		return indexOf(chr, 0);
	}
	// w̕ǂ̈ʒuɂ邩ẅʒuTB
	// :
	//	chr TB
	//	from Tn߂ʒuB
	// Ԓl:
	//	̌CfbNXBȂ-1B
	int indexOf(wchar_t chr, size_t from)const {
		if (from >= length())
			return -1;
		const wchar_t* found = wcschr(string + from, chr);
		if (found == NULL)
			return -1;
		return found - string;
	}
	// w̕񂪂ǂ̈ʒuɂ邩TB
	// :
	//	str TB
	// Ԓl:
	//	̌CfbNXBȂ-1B
	int indexOf(const wchar_t* str)const {
		return indexOf(str, 0);
	}
	// w̕񂪂ǂ̈ʒuɂ邩ẅʒuTB
	// :
	//	str TB
	//	from Tn߂ʒuB
	// Ԓl:
	//	̌CfbNXBȂ-1B
	//
	int indexOf(const wchar_t* str, size_t from)const {
		if (from >= length())
			return -1;
		const wchar_t* found = wcsstr(string + from, str);
		if (found == NULL)
			return -1;
		return found - string;
	}
	// ̒ԂB
	size_t length()const {
		return wcslen(string);
	}
	// w̕ŌɌʒu擾B
	// :
	//	chr TB
	// Ԓl:
	//	̌CfbNXBȂ-1B
	int lastIndexOf(char chr)const {
		return lastIndexOf(chr, (size_t) -1);
	}
	// w̕ẅʒuOōŌɌʒu擾B
	// :
	//	chr TB
	//	from Tn߂ʒuB
	// Ԓl:
	//	̌CfbNXBȂ-1B
	int lastIndexOf(wchar_t chr, size_t from)const {
		size_t len = length();
		if (from > len - 1)
			from = len - 1;
		const wchar_t* s = string;
		const wchar_t* end = string + from;
		const wchar_t* found = NULL;
		while (*s != '0' && s <= end) {
			if (*s == chr)
				found = s;
			s++;
		}
		return found != NULL ? found - string : -1;
	}
	// w̕񂪍ŌɌʒu擾B
	// :
	//	str TB
	// Ԓl:
	//	̌CfbNXBȂ-1B
	int lastIndexOf(const wchar_t* str)const {
		return lastIndexOf(str, (size_t) -1);
	}
	// w̕񂪎ẅʒuOōŌɌʒu擾B
	// :
	//	str TB
	//	from Tn߂ʒuB
	// Ԓl:
	//	̌CfbNXBȂ-1B
	int lastIndexOf(const wchar_t* str, size_t from)const {
		size_t len = length();
		size_t str_len = wcslen(str);
		if (from > len - str_len)
			from = len - str_len;
		const wchar_t* s = string + from;
		while (s >= string) {
			if (wcsncmp(s, str, str_len) == 0)
				return s - string;
			s--;
		}
		return -1;
	}
	// ̈ꕔoB
	// :
	//	start o̐擪̈ʒuB
	// Ԓl:
	//	̈ꕔB
	WString substring(int start)const {
		return WString(string + start);
	}
	// ̈ꕔoB
	// :
	//	start o̐擪̈ʒuB
	//	end ǒ̈ʒuB
	// Ԓl:
	//	̈ꕔB
	WString substring(int start, int end)const {
		return WString(string + start, end - start);
	}
	// ẅʒuɂ镶oB
	// :
	//	index öʒuB
	// Ԓl:
	//	ẅʒuɂ镶B
	char charAt(size_t index)const {
		return index < length() ? string[index] : '\0';
	}
	// w̕w̕ɒu܂B
	// :
	//	oldChr ̕B
	//	newChr u镶B
	// Ԓl:
	//	u̕B
	WString replace(char oldChr, char newChr)const {
		WString result(string);
		char* s = (char*) result.string;
		while (*s != '\0'){
			if (WString::isLeadByte(*s))
				s++;
			else if (*s == oldChr)
				*s = newChr;
			s++;
		}
		return result;
	}
	// 񒆂̑啶ɕϊB
	// Ԓl:
	//	ϊ̕B
	WString toLowerCase()const {
		WString result(string);
		char* s = (char*) result.string;
		while (*s != '\0'){
			if (WString::isLeadByte(*s))
				s++;
			else if ('A' <= *s && *s <= 'Z')
				*s += 'a' - 'A';
			s++;
		}
		return result;
	}
	// 񒆂̏啶ɕϊB
	// Ԓl:
	//	ϊ̕B
	WString toUpperCase()const {
		WString result(string);
		char* s = (char*) result.string;
		while (*s != '\0'){
			if (WString::isLeadByte(*s))
				s++;
			else if ('a' <= *s && *s <= 'z')
				*s += 'A' - 'a';
			s++;
		}
		return result;
	}
	// ̑Ő󔒕폜B
	// Ԓl:
	//	폜̕B
	WString trim()const {
		const wchar_t* s = string;
		while (*s != '\0' && (unsigned char) *s <= ' ')
			s++;
		const wchar_t* start = s;
		s = string + length();
		while (s > start && (*s != '\0' && (unsigned char) *s <= ' '))
			s--;
		return WString(start, s - start);
	}

	// operators

	// const char*ւ̃LXgZq
	// Ԓl:
	//	ւ̃AhXB
	operator const wchar_t*()const {
		return string;
	}
	// charẑ悤Ɉ߂[]ZqB
	// :
	//	index 擾镶̃CfbNXB
	// Ԓl:
	//	w̃CfbNXɂ镶B
	char operator[](size_t index)const {
		return charAt(index);
	}
	// A邽߂+ZqB
	// :
	//	source A镶B
	// Ԓl:
	//	AB
	WString operator+(const wchar_t* source)const {
		return WString(string, source);
	}
	// A邽߂+ZqB
	// :
	//	source A镶B
	// Ԓl:
	//	AB
	WString operator+(const WString& source)const {
		return *string != '\0' ? WString(string, source.string) : source;
	}
	// A邽߂+ZqB
	// :
	//	str1 A镶(O)B
	//	str2 A镶()B
	// Ԓl:
	//	AB
	friend WString operator+(const wchar_t* str1, const WString& str2) {
		return *str1 != '\0' ? WString(str1, str2.string) : str2;
	}
	// ZqB
	// :
	//	source 镶B
	// Ԓl:
	//	ʁB
	WString& operator=(const wchar_t* source) {
		set(create(source));
		return *this;
	}
	// ZqB
	// :
	//	source 镶B
	// Ԓl:
	//	ʁB
	WString& operator=(const WString& source) {
		set(source.string);
		return *this;
	}
	// Aʂ鉉ZqB
	// :
	//	source A镶B
	// Ԓl:
	//	AʁB
	WString& operator+=(const wchar_t* source) {
		if (*source != '\0')
			set(concat(string, source));
		return *this;
	}
	// rZqB
	// :
	//	str rΏۂ̕B
	// Ԓl:
	//	str̕ΐ^B
	bool operator==(const WString& str)const {
		return compareTo(str.string) == 0;
	}
	// rZqB
	// :
	//	str rΏۂ̕B
	// Ԓl:
	//	str̕ΐ^B
	bool operator==(const wchar_t* str)const {
		return compareTo(str) == 0;
	}
	// rZqB
	// :
	//	str1 r镶B
	//	str2 r镶B
	// Ԓl:
	//	str1str2̕ΐ^B
	friend bool operator==(const wchar_t* str1, const WString& str2) {
		return str2.compareTo(str1) == 0;
	}
	// rZqB
	// :
	//	str rΏۂ̕B
	// Ԓl:
	//	str̕Ȃΐ^B
	bool operator!=(const WString& str)const {
		return compareTo(str) != 0;
	}
	// rZqB
	// :
	//	str rΏۂ̕B
	// Ԓl:
	//	str̕Ȃΐ^B
	bool operator!=(const wchar_t* str)const {
		return compareTo(str) != 0;
	}
	// rZqB
	// :
	//	str1 r镶B
	//	str2 r镶B
	// Ԓl:
	//	str1str2̕Ȃΐ^B
	friend bool operator!=(const wchar_t* str1, const WString& str2) {
		return str2.compareTo(str1) != 0;
	}
	// rZqB
	// :
	//	str rΏۂ̕B
	// Ԓl:
	//	str̕傫ΐ^B
	bool operator<(const WString& str)const {
		return compareTo(str) < 0;
	}
	// rZqB
	// :
	//	str rΏۂ̕B
	// Ԓl:
	//	str̕傫ΐ^B
	bool operator<(const wchar_t* str)const {
		return compareTo(str) < 0;
	}
	// rZqB
	// :
	//	str1 r镶B
	//	str2 r镶B
	// Ԓl:
	//	str1str2̕傫ΐ^B
	friend bool operator<(const wchar_t* str1, const WString& str2) {
		return str2.compareTo(str1) > 0;
	}
	// rZqB
	// :
	//	str rΏۂ̕B
	// Ԓl:
	//	str̕傫ΐ^B
	bool operator<=(const WString& str)const {
		return compareTo(str) <= 0;
	}
	// rZqB
	// :
	//	str rΏۂ̕B
	// Ԓl:
	//	str̕傫ΐ^B
	bool operator<=(const wchar_t* str)const {
		return compareTo(str) <= 0;
	}
	// rZqB
	// :
	//	str1 r镶B
	//	str2 r镶B
	// Ԓl:
	//	str1str2̕傫ΐ^B
	friend bool operator<=(const wchar_t* str1, const WString& str2) {
		return str2.compareTo(str1) >= 0;
	}
	// rZqB
	// :
	//	str rΏۂ̕B
	// Ԓl:
	//	str̕ΐ^B
	bool operator>(const WString& str)const {
		return compareTo(str) > 0;
	}
	// rZqB
	// :
	//	str rΏۂ̕B
	// Ԓl:
	//	str̕ΐ^B
	bool operator>(const wchar_t* str)const {
		return compareTo(str) > 0;
	}
	// rZqB
	// :
	//	str1 r镶B
	//	str2 r镶B
	// Ԓl:
	//	str1str2̕ΐ^B
	friend bool operator>(const wchar_t* str1, const WString& str2) {
		return str2.compareTo(str1) < 0;
	}
	// rZqB
	// :
	//	str rΏۂ̕B
	// Ԓl:
	//	str̕ΐ^B
	bool operator>=(const WString& str)const {
		return compareTo(str) >= 0;
	}
	// rZqB
	// :
	//	str rΏۂ̕B
	// Ԓl:
	//	str̕ΐ^B
	bool operator>=(const wchar_t* str)const {
		return compareTo(str) >= 0;
	}
	// rZqB
	// :
	//	str1 r镶B
	//	str2 r镶B
	// Ԓl:
	//	str1str2̕ΐ^B
	friend bool operator>=(const wchar_t* str1, const WString& str2) {
		return str2.compareTo(str1) <= 0;
	}

	// public utilities

	// 2oCg̍ŏ1oCgǂ𔻒肷B
	// :
	//	肷oCgB
	// Ԓl:
	//	2oCg̍ŏ1oCgłΐ^B
	static bool isLeadByte(char ch) {
	#ifdef _INC_WINDOWS
		return ::IsDBCSLeadByte(ch) != 0;
	#else
		return (ch & 0x80) != 0;
	#endif
	}
};

}

#endif//_YCL_WSTRING_H_
