/*
 * Copyright (C) 2021- TeraTerm Project
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* StringBuffer.h ɍ쐬 */

#ifndef _YCL_WSTRINGBUFFER_H_
#define _YCL_WSTRINGBUFFER_H_

#pragma once

#include <YCL/common.h>

#include <YCL/WString.h>

#include <malloc.h>

namespace yebisuya {

// ϒ̕߂̃NXB
class WStringBuffer {
private:
	// i[obt@B
	wchar_t* buffer;
	// ݗLȕ̒B
	size_t validLength;
	// obt@̑傫B
	size_t bufferSize;
	enum {
		// obt@LۂɎgpTCYB
		INIT_CAPACITY = 16,
	};
	// obt@B
	// :
	//	source	B
	//	length	̒B
	//	capacity	obt@̏TCYB
	void init(const wchar_t* source, size_t length, size_t capacity) {
		if ((capacity != 0 || length != 0) && capacity < length + INIT_CAPACITY)
			capacity = length + INIT_CAPACITY;
		validLength = length;
		bufferSize = capacity;
		if (bufferSize == 0) {
			buffer = NULL;
		}else{
			buffer = new wchar_t[bufferSize];
		}
		if (source != NULL) {
			memcpy(buffer, source, validLength);
		}
		memset(buffer + validLength, '\0', bufferSize - validLength);
	}
public:
	// ftHgRXgN^B
	WStringBuffer() {
		init(NULL, 0, 0);
	}
	// obt@̏TCYw肷RXgN^B
	// :
	//	capacity obt@̏TCYB
	WStringBuffer(size_t capacity) {
		init(NULL, 0, capacity);
	}
	// obt@̏w肷RXgN^B
	// :
	//	source	B
	WStringBuffer(const wchar_t* source) {
		init(source, wcslen(source), 0);
	}
	// Rs[RXgN^B
	// :
	//	source	B
	WStringBuffer(const WStringBuffer& source) {
		init(source.buffer, source.validLength, source.bufferSize);
	}
	// fXgN^B
	~WStringBuffer() {
		delete[] buffer;
	}

	// ݗLȕ̒擾B
	// Ԓl:
	//	Lȕ̒B
	size_t length()const {
		return validLength;
	}
	// obt@̃TCY擾B
	// Ԓl:
	//	obt@̃TCYB
	size_t capacity()const {
		return bufferSize;
	}
	// obt@̃TCYw̒܂悤ɒ߂B
	// :
	//	newLength	߂钷B
	void ensureCapacity(size_t newLength) {
		if (bufferSize < newLength) {
			wchar_t* oldBuffer = buffer;
			init(oldBuffer, validLength, newLength + INIT_CAPACITY);
			delete[] oldBuffer;
		}
	}
	// Lȕ񒷂ύXB
	// :
	//	newLength	V񒷁B
	void setLength(size_t newLength) {
		if (validLength < newLength)
			ensureCapacity(newLength);
		validLength = newLength;
	}
	// ẅʒu̕擾B
	// :
	//	index	̈ʒuB
	// Ԓl:
	//	ẅʒu̕B
	char charAt(size_t index)const {
		return index < validLength ? buffer[index] : '\0';
	}
	// ẅʒu̕擾B
	// :
	//	index	̈ʒuB
	// Ԓl:
	//	ẅʒu̎̕QƁB
	wchar_t& charAt(size_t index) {
		if (index >= validLength) {
			ensureCapacity(validLength + 1);
			index = validLength++;
		}
		return buffer[index];
	}
	// ẅʒu̕ύXB
	// :
	//	index	ύX镶̈ʒuB
	//	chr	ύX镶B
	void setCharAt(int index, char chr) {
		charAt(index) = chr;
	}
	// ǉB
	// :
	//	chr	ǉ镶B
	// Ԓl:
	//	ǉʁB
	WStringBuffer& append(wchar_t chr) {
		charAt(validLength) = chr;
		return *this;
	}
	// ǉB
	// :
	//	source	ǉ镶B
	// Ԓl:
	//	ǉʁB
	WStringBuffer& append(const wchar_t* source) {
		return append(source, wcslen(source));
	}
	// ǉB
	// :
	//	source	ǉ镶B
	//	length	ǉ镶̒B
	// Ԓl:
	//	ǉʁB
	WStringBuffer& append(const wchar_t* source, size_t length) {
		size_t oldLength = validLength;
		ensureCapacity(validLength + length);
		memcpy(buffer + oldLength, source, length);
		validLength += length;
		return *this;
	}
	// ẅʒu̕폜B
	// :
	//	start	폜ʒuB
	// Ԓl:
	//	폜ʁB
	WStringBuffer& remove(size_t index) {
		return remove(index, index + 1);
	}
	// ẅʒu̕폜B
	// :
	//	start	폜擪ʒuB
	//	end	폜ÏʒuB
	// Ԓl:
	//	폜ʁB
	WStringBuffer& remove(size_t start, size_t end) {
		if (start < end) {
			if (end < validLength){
				memcpy(buffer + start, buffer + end, validLength - end);
				validLength -= end - start;
			}else{
				validLength = start;
			}
		}
		return *this;
	}
	// ẅʒu̕uB
	// :
	//	start	u擪ʒuB
	//	end	uÏʒuB
	//	source	u镶B
	// Ԓl:
	//	uʁB
	WStringBuffer& replace(size_t start, size_t end, const char* source) {
		if (end > validLength)
			end = validLength;
		if (start < end) {
			size_t length = strlen(source);
			size_t oldLength = validLength;
			ensureCapacity(validLength += length - (end - start));
			memcpy(buffer + start + length, buffer + end, oldLength - end);
			memcpy(buffer + start, source, length);
		}
		return *this;
	}
	// ẅʒu̕擾B
	// :
	//	start	擾镶̐擪ʒuB
	// Ԓl:
	//	ẅʒu̕B
	WString substring(size_t index)const {
		return WString(buffer + index, validLength - index);
	}
	// ẅʒu̕擾B
	// :
	//	start	擾镶̐擪ʒuB
	//	end	擾镶̏ÏʒuB
	// Ԓl:
	//	ẅʒu̕B
	WString substring(size_t start, size_t end)const {
		if (end > validLength)
			end = validLength;
		return WString(buffer + start, end - start);
	}
	// ẅʒuɕ}B
	// :
	//	index	}ʒuB
	//	source	}镶B
	// Ԓl:
	//	}ʁB
	WStringBuffer& insert(size_t index, char chr) {
		return insert(index, &chr, 1);
	}
	// ẅʒuɕ}B
	// :
	//	index	}ʒuB
	//	source	}镶B
	// Ԓl:
	//	}ʁB
	WStringBuffer& insert(size_t index, const char* source) {
		return insert(index, source, strlen(source));
	}
	// ẅʒuɕ}B
	// :
	//	index	}ʒuB
	//	source	}镶B
	//	length	̒B
	// Ԓl:
	//	}ʁB
	WStringBuffer& insert(size_t index, const char* source, size_t length) {
		if (index >= validLength)
			index = validLength;
		size_t oldLength = validLength;
		ensureCapacity(validLength + length);
		char* temp = (char*) alloca(oldLength - index);
		memcpy(temp, buffer + index, oldLength - index);
		memcpy(buffer + index, source, length);
		memcpy(buffer + index + length, temp, oldLength - index);
		validLength += length;
		return *this;
	}
#if 0
	// 𔽓]B
	// Ԓl:
	//	]ʁB
	WStringBuffer& reverse() {
		char* temporary = (char*) alloca(sizeof (char) * validLength);
		char* dst = temporary + validLength;
		wchar_t* src = buffer;
		while (temporary < dst) {
			if (String::isLeadByte(*src)) {
				char pre = *src++;
				*--dst = *src++;
				*--dst = pre;
			}else{
				*--dst = *src++;
			}
		}
		memcpy(buffer, temporary, validLength);
		return *this;
	}
#endif
	// 擾B
	// Ԓl:
	//	ݐݒ肳Ă镶B
	WString toString()const {
		return WString(buffer, validLength);
	}

	// ꕶ̕ɕύXB
	// :
	//	ύXꕶB
	// Ԓl:
	//	ύXʁB
	WStringBuffer& set(char chr) {
		ensureCapacity(1);
		buffer[0] = chr;
		validLength = 1;
		return *this;
	}
	// w̕ɕύXB
	// :
	//	source	ύX镶B
	// Ԓl:
	//	ύXʁB
	WStringBuffer& set(const char* source) {
		size_t length = strlen(source);
		ensureCapacity(validLength = length);
		memcpy(buffer, source, length);
		return *this;
	}

	// char*ɕϊLXgZqB
	// obt@̃AhX擾B
	// Ԓl:
	//	obt@̃AhXB
	operator wchar_t*() {
		return buffer;
	}
	// StringɕϊLXgZqB
	// 擾B
	// Ԓl:
	//	ݐݒ肳Ă镶B
	operator WString()const {
		return toString();
	}
	// ZqB
	// ꕶ̕ɕύXB
	// :
	//	ch	ύXꕶB
	// Ԓl:
	//	ʁB
	WStringBuffer& operator=(char ch) {
		return set(ch);
	}
	// ZqB
	// w̕ɕύXB
	// :
	//	source	ύX镶B
	// Ԓl:
	//	ʁB
	WStringBuffer& operator=(const char* source) {
		return set(source);
	}
	// AZqB
	// ǉB
	// :
	//	ch	ǉ镶B
	// Ԓl:
	//	ʁB
	WStringBuffer& operator+=(char ch) {
		return append(ch);
	}
	// AZqB
	// ǉB
	// :
	//	source	ǉ镶B
	// Ԓl:
	//	ʁB
	WStringBuffer& operator+=(const wchar_t* source) {
		return append(source);
	}
};

}

#endif//_YCL_WSTRINGBUFFER_H_
