/*
 * (C) 2019- TeraTerm Project
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <windows.h>
#include <assert.h>
#if !defined(_CRTDBG_MAP_ALLOC)
#define _CRTDBG_MAP_ALLOC
#endif
#include <crtdbg.h>

#include "compat_win.h"
#include "ttlib.h"	// for IsWindowsXPOrLater()

#include "dllutil.h"

typedef struct {
	const wchar_t *fname;
	BOOL NeedFreeLibrary;
	HMODULE handle;
	int refCount;
} HandleList_t;

static HandleList_t *HandleList;
static int HandleListCount;

static HMODULE GetHandle(const wchar_t *fname, DLLLoadFlag LoadFlag)
{
	HMODULE module;
	HandleList_t *p;

	// ȑOɃ[h?
	if (LoadFlag != DLL_GET_MODULE_HANDLE) {
		p = HandleList;
		for (int i = 0; i < HandleListCount; i++) {
			if (wcscmp(p->fname, fname) == 0) {
				p->refCount++;
				return p->handle;
			}
			p++;
		}
	}

	// W[[hς݂Ȃnh擾ł
	// exe[hƂdll[hĂ΁AnhԂĂ
	module = GetModuleHandleW(fname);
	if (LoadFlag == DLL_GET_MODULE_HANDLE) {
		// module == NULL ł return 
		return module;
	}

	BOOL NeedFreeLibrary = FALSE;
	if (module == NULL) {
		// VɃ[h
		int r;
		wchar_t dllPath[MAX_PATH];
		dllPath[0] = 0;
		switch (LoadFlag) {
		case DLL_LOAD_LIBRARY_SxS:
			if (IsWindowsXPOrLater()) {
				// Side by Side [fBO𗘗p邽߃tpXɂȂ
				;
			} else {
				// dllCWFNV΍ŃtpXŃ[h
				r = GetSystemDirectoryW(dllPath, _countof(dllPath));
				assert(r != 0);
				if (r == 0) return NULL;
				wcscat_s(dllPath, _countof(dllPath), L"\\");
			}
			break;
		case DLL_LOAD_LIBRARY_SYSTEM:
			r = GetSystemDirectoryW(dllPath, _countof(dllPath));
			assert(r != 0);
			if (r == 0) return NULL;
			wcscat_s(dllPath, _countof(dllPath), L"\\");
			break;
		case DLL_LOAD_LIBRARY_CURRENT:
			r = GetModuleFileNameW(NULL, dllPath, _countof(dllPath));
			assert(r != 0);
			if (r == 0) return NULL;
			*(wcsrchr(dllPath, L'\\') + 1) = 0;
			break;
		default:
			return NULL;
		}
		wcscat_s(dllPath, _countof(dllPath), fname);
		module = LoadLibraryW(dllPath);
		if (module == NULL) {
			// ݂Ȃ,dllȂ?
			return NULL;
		}
		NeedFreeLibrary = TRUE;
	}

	// Xgɒǉ
	HandleList = (HandleList_t *)realloc(HandleList, sizeof(HandleList_t) * (HandleListCount + 1));
	p = &HandleList[HandleListCount];
	p->fname = _wcsdup(fname);
	p->handle = module;
	p->NeedFreeLibrary = NeedFreeLibrary;
	p->refCount = 1;
	HandleListCount++;
	return module;
}

static void DLLFree(HandleList_t *p)
{
	if (p->NeedFreeLibrary) {
		FreeLibrary(p->handle);
	}
	free((void *)p->fname);
}

static void DLLFreeFromList(int no)
{
	HandleList_t *p = &HandleList[no];
	p->refCount--;
	if (p->refCount > 0) {
		return;
	}

	// J
	DLLFree(p);
	memcpy(p, p+1, sizeof(*p) * (HandleListCount - no - 1));
	HandleListCount--;
	HandleList = (HandleList_t *)realloc(HandleList, sizeof(HandleList_t)*HandleListCount);
}

/**
 *	dllA[h
 *
 *	@param[in]	handle
 */
void DLLFreeByHandle(HANDLE handle)
{
	// XgT
	HandleList_t *p = HandleList;
	for (int i = 0; i < HandleListCount; i++) {
		if (p->handle == handle) {
			// A폜
			DLLFreeFromList(i);
			return;
		}
		p++;
	}

	// XgɂȂ
}

/**
 *	dllA[h
 *
 *	@param[in]	fname	t@C
 */
void DLLFreeByFileName(const wchar_t *fname)
{
	// XgT
	HandleList_t *p = HandleList;
	for (int i = 0; i < HandleListCount; i++) {
		if (wcscmp(p->fname, fname) == 0) {
			// A폜
			DLLFreeFromList(i);
			return;
		}
		p++;
	}

	// XgɂȂ
}

/**
 * DLL̊֐ւ̃AhX擾
 * @param[in]		fname		DLLt@C
 * @param[in] 		FuncFlag	֐ȂƂ̓
 *					DLL_ACCEPT_NOT_EXIST	ȂĂok
 *					DLL_ERROR_NOT_EXIST		ȂꍇG[
 * @param[in]		ApiName		API
 * @param[in,out]	pFunc		֐ւ̃AhX
 *								ȂNULL
 * @retval	NO_ERROR				G[Ȃ
 * @retval	ERROR_FILE_NOT_FOUND	DLLȂ(sȃt@C)
 * @retval	ERROR_PROC_NOT_FOUND	֐GgȂ
 */
DWORD DLLGetApiAddress(const wchar_t *fname, DLLLoadFlag LoadFlag,
					   const char *ApiName, void **pFunc)
{
	HMODULE hDll = GetHandle(fname, LoadFlag);
	if (hDll == NULL) {
		*pFunc = NULL;
		return ERROR_FILE_NOT_FOUND;
	} else {
		*pFunc = (void *)GetProcAddress(hDll, ApiName);
		if (*pFunc == NULL) {
			return ERROR_PROC_NOT_FOUND;
		}
		return NO_ERROR; // = 0
	}
}

/**
 * DLL̊̕֐ւ̃AhX擾
 * @param[in] fname		dll̃t@C
 * @param[in] LoadFlag	dll̃[h
 *				DLL_GET_MODULE_HANDLE,
 *				DLL_LOAD_LIBRARY_SYSTEM,
 *				DLL_LOAD_LIBRARY_CURRENT,
 * @param[in] FuncFlag	֐ȂƂ̓
 * @param[out] handle	DLLnh
 *				NULL̂Ƃ͒lԂȂ
 * @retval	NO_ERROR				G[Ȃ
 * @retval	ERROR_FILE_NOT_FOUND	DLLȂ(sȃt@C)
 * @retval	ERROR_PROC_NOT_FOUND	֐GgȂ
 */
DWORD DLLGetApiAddressFromList(const wchar_t *fname, DLLLoadFlag LoadFlag,
							   DLLFuncFlag FuncFlag, const APIInfo *ApiInfo, HANDLE *handle)
{
	HMODULE hDll = GetHandle(fname, LoadFlag);
	if (hDll == NULL) {
		while(ApiInfo->ApiName != NULL) {
			void **func = ApiInfo->func;
			*func = NULL;
			ApiInfo++;
		}
		if (handle != NULL) {
			*handle = NULL;
		}
		return ERROR_FILE_NOT_FOUND;
	} else {
		BOOL exist_all = TRUE;
		const APIInfo *p = ApiInfo;

		// AhX擾
		while(p->ApiName != NULL) {
			void **func = p->func;
			*func = (void *)GetProcAddress(hDll, p->ApiName);
			if (*func == NULL) {
				exist_all = FALSE;
			}
			p++;
		}

		// ׂČ or ȂAPIĂok
		if (exist_all || FuncFlag == DLL_ACCEPT_NOT_EXIST) {
			if (handle != NULL) {
				*handle = hDll;
			}
			return NO_ERROR;
		}

		// ȂAPÎŃG[
		p = ApiInfo;
		while(p->ApiName != NULL) {
			void **func = p->func;
			*func = NULL;
			p++;
		}
		DLLFreeByFileName(fname);
		if (handle != NULL) {
			*handle = NULL;
		}
		return ERROR_PROC_NOT_FOUND;
	}
}

void DLLGetApiAddressFromLists(const DllInfo *dllInfos)
{
	while (dllInfos->DllName != NULL) {
		DLLGetApiAddressFromList(dllInfos->DllName,
								 dllInfos->LoadFlag,
								 dllInfos->FuncFlag,
								 dllInfos->APIInfoPtr, NULL);
		dllInfos++;
	}
}

static void SetupLoadLibraryPath(void)
{
	const wchar_t *kernel32 = L"kernel32.dll";

	// SetDefaultDllDirectories() gꍇ́A
	// pX %WINDOWS%\system32 ݂̂ɐݒ肷
	DLLGetApiAddress(kernel32, DLL_GET_MODULE_HANDLE,
					 "SetDefaultDllDirectories", (void **)&pSetDefaultDllDirectories);
	if (pSetDefaultDllDirectories != NULL) {
		pSetDefaultDllDirectories(LOAD_LIBRARY_SEARCH_SYSTEM32);
		return;
	}

	// SetDefaultDllDirectories() gȂĂ
	// SetDllDirectory() gꍇ
	// JgfBNgłpX͂ĂB
	// JgO(""Zbg)Ȃ̂ŁAANSIłok
	DLLGetApiAddress(kernel32, DLL_GET_MODULE_HANDLE,
					 "SetDllDirectoryA", (void **)&pSetDllDirectoryA);
	if (pSetDllDirectoryA != NULL) {
		pSetDllDirectoryA("");
	}
}

void DLLInit()
{
	if (HandleListCount != 0) {
		return;		// ς
	}
	HandleList = NULL;
	HandleListCount = 0;
	SetupLoadLibraryPath();
}

void DLLExit()
{
	if (HandleListCount == 0) {
		return;		// gp
	}
	for (int i = 0; i < HandleListCount; i++) {
		HandleList_t *p = &HandleList[i];
		DLLFree(p);
	}
	free(HandleList);
	HandleList = NULL;
	HandleListCount = 0;
}
