/*
 * Copyright (C) 1994-1998 T. Teranishi
 * (C) 2006- TeraTerm Project
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* misc. routines  */

#include <sys/stat.h>
#include <sys/utime.h>
#include <string.h>
#include <time.h>
#include <stdio.h>
#include <shlobj.h>
#include <ctype.h>
#include <assert.h>

#include "teraterm.h"
#include "tttypes.h"
#include "compat_win.h"

#include "../teraterm/unicode_test.h"


/**
 *	EBhËʒu擾
 *	@Param[in]		hWnd
 *	@Param[in]		point		ʒu(x,y)
 *	@Param[in,out]	InWindow	EBhE
 *	@Param[in,out]	InClient	NCAg̈
 *	@Param[in,out]	InTitleBar	^Cgo[
 *	@retval			FALSE		hWnd
 */
BOOL GetPositionOnWindow(
	HWND hWnd, const POINT *point,
	BOOL *InWindow, BOOL *InClient, BOOL *InTitleBar)
{
	const int x = point->x;
	const int y = point->y;
	RECT winRect;
	RECT clientRect;

	if (InWindow != NULL) *InWindow = FALSE;
	if (InClient != NULL) *InClient = FALSE;
	if (InTitleBar != NULL) *InTitleBar = FALSE;

	if (!GetWindowRect(hWnd, &winRect)) {
		return FALSE;
	}

	if ((x < winRect.left) || (winRect.right < x) ||
		(y < winRect.top) || (winRect.bottom < y))
	{
		return TRUE;
	}
	if (InWindow != NULL) *InWindow = TRUE;

	{
		POINT pos;
		GetClientRect(hWnd, &clientRect);
		pos.x = clientRect.left;
		pos.y = clientRect.top;
		ClientToScreen(hWnd, &pos);
		clientRect.left = pos.x;
		clientRect.top = pos.y;

		pos.x = clientRect.right;
		pos.y = clientRect.bottom;
		ClientToScreen(hWnd, &pos);
		clientRect.right = pos.x;
		clientRect.bottom = pos.y;
	}

	if ((clientRect.left <= x) && (x < clientRect.right) &&
		(clientRect.top <= y) && (y < clientRect.bottom))
	{
		if (InClient != NULL) *InClient = TRUE;
		if (InTitleBar != NULL) *InTitleBar = FALSE;
		return TRUE;
	}
	if (InClient != NULL) *InClient = FALSE;

	if (InTitleBar != NULL) {
		*InTitleBar = (y < clientRect.top) ? TRUE : FALSE;
	}

	return TRUE;
}

int SetDlgTextsW(HWND hDlgWnd, const DlgTextInfo *infos, int infoCount, const wchar_t *UILanguageFile)
{
	return SetI18nDlgStrsW(hDlgWnd, "Tera Term", infos, infoCount, UILanguageFile);
}

int SetDlgTexts(HWND hDlgWnd, const DlgTextInfo *infos, int infoCount, const char *UILanguageFile)
{
	return SetI18nDlgStrsA(hDlgWnd, "Tera Term", infos, infoCount, UILanguageFile);
}

void SetDlgMenuTextsW(HMENU hMenu, const DlgTextInfo *infos, int infoCount, const wchar_t *UILanguageFile)
{
	SetI18nMenuStrsW(hMenu, "Tera Term", infos, infoCount, UILanguageFile);
}

void SetDlgMenuTexts(HMENU hMenu, const DlgTextInfo *infos, int infoCount, const char *UILanguageFile)
{
	SetI18nMenuStrs("Tera Term", hMenu, infos, infoCount, UILanguageFile);
}

/**
 *	EBhE\ĂfBXvC̃fXNgbv͈̔͂擾
 *	@param[in]		hWnd	EBhẼnh
 *	@param[out]		rect	fXNgbv
 */
void GetDesktopRect(HWND hWnd, RECT *rect)
{
	if (pMonitorFromWindow != NULL) {
		// }`j^T|[gĂꍇ
		MONITORINFO monitorInfo;
		HMONITOR hMonitor = pMonitorFromWindow(hWnd, MONITOR_DEFAULTTONEAREST);
		monitorInfo.cbSize = sizeof(MONITORINFO);
		pGetMonitorInfoA(hMonitor, &monitorInfo);
		*rect = monitorInfo.rcWork;
	} else {
		// }`j^T|[gĂȂꍇ
		SystemParametersInfo(SPI_GETWORKAREA, 0, rect, 0);
	}
}

/**
 *	point݂fBXvC̃fXNgbv͈̔͂擾
 */
static void GetDesktopRectFromPoint(const POINT *p, RECT *rect)
{
	if (pMonitorFromPoint == NULL) {
		// NT4.0, 95 ̓}`j^APIɔΉ
		SystemParametersInfo(SPI_GETWORKAREA, 0, rect, 0);
	}
	else {
		// }`j^T|[gĂꍇ
		HMONITOR hm;
		POINT pt;
		MONITORINFO mi;

		pt.x = p->x;
		pt.y = p->y;
		hm = pMonitorFromPoint(pt, MONITOR_DEFAULTTONEAREST);

		mi.cbSize = sizeof(MONITORINFO);
		pGetMonitorInfoA(hm, &mi);
		*rect = mi.rcWork;
	}
}

/**
 *	EBhË悩͂ݏoȂ悤Ɉړ
 *	͂ݏoĂȂꍇ͈ړȂ
 *
 *	@param[in]	hWnd		ʒu𒲐EBhE
 */
void MoveWindowToDisplayRect(HWND hWnd, const RECT *rect)
{
	RECT desktop = *rect;
	RECT win_rect;
	int win_width;
	int win_height;
	int win_x;
	int win_y;
	BOOL modify = FALSE;

	GetDesktopRect(hWnd, &desktop);

	GetWindowRect(hWnd, &win_rect);
	win_x = win_rect.left;
	win_y = win_rect.top;
	win_height = win_rect.bottom - win_rect.top;
	win_width = win_rect.right - win_rect.left;
	if (win_y < desktop.top) {
		win_y = desktop.top;
		modify = TRUE;
	}
	else if (win_y + win_height > desktop.bottom) {
		win_y = desktop.bottom - win_height;
		modify = TRUE;
	}
	if (win_x < desktop.left) {
		win_x = desktop.left;
		modify = TRUE;
	}
	else if (win_x + win_width > desktop.right) {
		win_x = desktop.right - win_width;
		modify = TRUE;
	}

	if (modify) {
		SetWindowPos(hWnd, NULL, win_x, win_y, 0, 0, SWP_NOSIZE | SWP_NOZORDER);
	}
}

/**
 *	EBhEfBXvC͂ݏoȂ悤Ɉړ
 *	fBXvC̓EBhEĂfBXvC
 *	܂łƂ͖ʐς̍L
 *	͂ݏoĂȂꍇ͈ړȂ
 *
 *	@param[in]	hWnd		ʒu𒲐EBhE
 */
void MoveWindowToDisplay(HWND hWnd)
{
	RECT desktop;
	GetDesktopRect(hWnd, &desktop);
	MoveWindowToDisplayRect(hWnd, &desktop);
}

/**
 *	EBhEfBXvC͂ݏoȂ悤Ɉړ
 *	fBXvCpointĂfBXvC
 *	͂ݏoĂȂꍇ͈ړȂ
 *
 *	@param[in]	hWnd		ʒu𒲐EBhE
 *	@param[in]	point		ʒu
 */
void MoveWindowToDisplayPoint(HWND hWnd, const POINT *point)
{
	RECT desktop;
	GetDesktopRectFromPoint(point, &desktop);
	MoveWindowToDisplayRect(hWnd, &desktop);
}

/**
 *	EBhEfBXvC̒ɔzu
 *
 *	@param[in]	hWnd		ʒu𒲐EBhE
 *	@param[in]	hWndParent	̃EBhE̒Ɉړ
 *							(NULL̏ꍇfBXvC̒)
 *
 *	hWndParent̎w肪ꍇ
 *		hWndParent\Ԃ̏ꍇ
 *			- hWndParent̒ɔzu
 *			- \ĂfBXvC͂ݏoꍇ͒
 *		hWndParent\Ԃ̏ꍇ
 *			- hWnd\ĂfBXvC̒ɔzu
 *	hWndParentNULL̏ꍇ
 *		- hWnd\ĂfBXvC̒ɔzu
 */
void CenterWindow(HWND hWnd, HWND hWndParent)
{
	RECT rcWnd;
	LONG WndWidth;
	LONG WndHeight;
	int NewX;
	int NewY;
	RECT rcDesktop;
	BOOL r;

	r = GetWindowRect(hWnd, &rcWnd);
	assert(r != FALSE); (void)r;
	WndWidth = rcWnd.right - rcWnd.left;
	WndHeight = rcWnd.bottom - rcWnd.top;

	if (hWndParent == NULL || !IsWindowVisible(hWndParent) || IsIconic(hWndParent)) {
		// eݒ肳ĂȂ or \ĂȂ or iconĂ ꍇ
		// EBhE̕\ĂfBXvC̒ɕ\
		GetDesktopRect(hWnd, &rcDesktop);

		// fXNgbv(\ĂfBXvC)̒
		NewX = (rcDesktop.left + rcDesktop.right) / 2 - WndWidth / 2;
		NewY = (rcDesktop.top + rcDesktop.bottom) / 2 - WndHeight / 2;
	} else {
		RECT rcParent;
		r = GetWindowRect(hWndParent, &rcParent);
		assert(r != FALSE); (void)r;

		// hWndParent̒
		NewX = (rcParent.left + rcParent.right) / 2 - WndWidth / 2;
		NewY = (rcParent.top + rcParent.bottom) / 2 - WndHeight / 2;

		GetDesktopRect(hWndParent, &rcDesktop);
	}

	// fXNgbv͂ݏoꍇA
	if (NewX + WndWidth > rcDesktop.right)
		NewX = rcDesktop.right - WndWidth;
	if (NewX < rcDesktop.left)
		NewX = rcDesktop.left;

	if (NewY + WndHeight > rcDesktop.bottom)
		NewY = rcDesktop.bottom - WndHeight;
	if (NewY < rcDesktop.top)
		NewY = rcDesktop.top;

	// ړ
	SetWindowPos(hWnd, NULL, NewX, NewY, 0, 0,
				 SWP_NOSIZE | SWP_NOZORDER | SWP_NOACTIVATE);
}

/**
 *	hWnd̕\Ă郂j^DPI擾
 *	Per-monitor DPI awarenessΉ
 *
 *	@param	hWnd	(NULL̂ƂPrimary monitor)
 *	@retval	DPIl(ʏDPI96)
 */
int GetMonitorDpiFromWindow(HWND hWnd)
{
	if (pGetDpiForMonitor == NULL) {
		// _CAOł͎XP[OĂ̂
		// 96Ԃ悤
		int dpiY;
		HDC hDC = GetDC(hWnd);
		dpiY = GetDeviceCaps(hDC,LOGPIXELSY);
		ReleaseDC(hWnd, hDC);
		return dpiY;
	} else {
		UINT dpiX;
		UINT dpiY;
		HMONITOR hMonitor;
		if (hWnd == NULL) {
			// https://devblogs.microsoft.com/oldnewthing/20070809-00/?p=25643
			const POINT ptZero = { 0, 0 };
			hMonitor = pMonitorFromPoint(ptZero, MONITOR_DEFAULTTOPRIMARY);
		}
		else {
			hMonitor = pMonitorFromWindow(hWnd, MONITOR_DEFAULTTONEAREST);
		}
		pGetDpiForMonitor(hMonitor, 0 /*0=MDT_EFFECTIVE_DPI*/, &dpiX, &dpiY);
		return (int)dpiY;
	}
}

void OutputDebugPrintfW(const wchar_t *fmt, ...)
{
	wchar_t tmp[1024];
	va_list arg;
	va_start(arg, fmt);
	_vsnwprintf_s(tmp, _countof(tmp), _TRUNCATE, fmt, arg);
	va_end(arg);
	OutputDebugStringW(tmp);
}

/* vim: set ts=4 sw=4 ff=dos : */
