/*
 * Copyright (C) 2008- TeraTerm Project
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/* original idea from PuTTY 0.60 windows/sizetip.c */

#include "teraterm.h"
#include "tttypes.h"
#include "ttlib.h"
#include "ttwinman.h"
#include "compat_win.h"

#include <windows.h>
#include <stdio.h>

#include "tipwin.h"

static TipWin *SizeTip;
static int tip_enabled = 0;

/*
 *	point 
 *	XN[͂ݏoĂꍇA悤ɕ␳
 *	NearestMonitor  TRUE ̂ƂAł߂j^
 *	FALSÊƂA}EX̂郂j^Ɉړ
 *	fBXvC̒[ FrameWidth(pixel) 藣悤ɂ
 */
static void FixPosFromFrame(POINT *point, int FrameWidth, BOOL NearestMonitor)
{
	if (HasMultiMonitorSupport()) {
		// }`j^T|[gĂꍇ
		HMONITOR hm;
		MONITORINFO mi;
		int ix, iy;

		// ̍WۑĂ
		ix = point->x;
		iy = point->y;

		hm = pMonitorFromPoint(*point, MONITOR_DEFAULTTONULL);
		if (hm == NULL) {
			if (NearestMonitor) {
				// ł߂j^ɕ\
				hm = pMonitorFromPoint(*point, MONITOR_DEFAULTTONEAREST);
			} else {
				// XN[͂ݏoĂꍇ̓}EX̂郂j^ɕ\
				GetCursorPos(point);
				hm = pMonitorFromPoint(*point, MONITOR_DEFAULTTONEAREST);
			}
		}

		mi.cbSize = sizeof(MONITORINFO);
		pGetMonitorInfoA(hm, &mi);
		if (ix < mi.rcMonitor.left + FrameWidth) {
			ix = mi.rcMonitor.left + FrameWidth;
		}
		if (iy < mi.rcMonitor.top + FrameWidth) {
			iy = mi.rcMonitor.top + FrameWidth;
		}
		
		point->x = ix;
		point->y = iy;
	}
	else
	{
		// }`j^T|[gĂȂꍇ
		if (point->x < FrameWidth) {
			point->x = FrameWidth;
		}
		if (point->y < FrameWidth) {
			point->y = FrameWidth;
		}
	}
}

/* TCYpc[`bv\
 *
 * F
 *   src        EBhEnh
 *   cx, cy     c[`bvɕ\cTCY
 *   fwSide     TCYɂǂ̃EBhE͂񂾂
 *   newX, newY TCY̍̍W
 *
 * : 
 * Windows9xł͉ʂ̃vpeBŁuEBhE̓e\܂
 * hbOvɃ`FbNȂƁAc[`bv\܂B
 */
void UpdateSizeTip(HWND src, int cx, int cy, UINT fwSide, int newX, int newY)
{
	char str[32];
	int tooltip_movable = 0;

	if (!tip_enabled)
		return;

	/* Generate the tip text */
	_snprintf_s(str, _countof(str), _TRUNCATE, "%dx%d", cx, cy);

	// EBhẺEAEA͂񂾏ꍇ́Ac[`bvɔzuB
	// ȊO̓TCY̍ɔzuB
	if (!(fwSide == WMSZ_RIGHT || fwSide == WMSZ_BOTTOMRIGHT || fwSide == WMSZ_BOTTOM)) {
		tooltip_movable = 1;
	}

	if (SizeTip == NULL) {
		RECT wr;
		POINT point;
		int w, h;

		// ̏cTCY擾
		TipWinGetTextWidthHeight(src, str, &w, &h);

		// EBhËʒu擾
		GetWindowRect(src, &wr);

		// sizetipoʒúAEBhE(X, Y)ɑ΂āA
		// (X, Y - ̍ - TIP_WIN_FRAME_WIDTH * 2) ƂB
		point.x = wr.left;
		point.y = wr.top - (h + TIP_WIN_FRAME_WIDTH * 2);
		FixPosFromFrame(&point, 16, FALSE);
		cx = point.x;
		cy = point.y;

		SizeTip = TipWinCreateA(NULL, src, cx, cy, str);
	} else {
		/* Tip already exists, just set the text */
		TipWinSetTextA(SizeTip, str);

		// EBhE̍オړꍇ
		if (tooltip_movable) {
			TipWinSetPos(SizeTip, newX + TIP_WIN_FRAME_WIDTH*2, newY + TIP_WIN_FRAME_WIDTH*2);
		}
	}
}

void EnableSizeTip(int bEnable)
{
	if (SizeTip && !bEnable) {
		TipWinDestroy(SizeTip);
		SizeTip = NULL;
	}

	tip_enabled = bEnable;
}
