/*
 * Copyright (c) 1998-2001, Robert O'Callahan
 * (C) 2004- TeraTerm Project
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
This code is copyright (C) 1998-1999 Robert O'Callahan.
See LICENSE.TXT for the license.
*/

#include "ttxssh.h"
#include "codeconv.h"

void UTIL_init_sock_write_buf(UTILSockWriteBuf *buf)
{
	buf_create(&buf->bufdata, &buf->buflen);
	buf->datastart = 0;
	buf->datalen = 0;
}

static int send_until_block(PTInstVar pvar, SOCKET s,
                            const char *data, int len)
{
	int total_sent = 0;

	while (len > 0) {
		int sent_amount = (pvar->Psend) (s, data, len, 0);

		if (sent_amount < 0) {
			if (WSAGetLastError() == WSAEWOULDBLOCK) {
				return total_sent;
			} else {
				return sent_amount;
			}
		} else {
			total_sent += sent_amount;
			data += sent_amount;
			len -= sent_amount;
		}
	}

	return total_sent;
}

/* Tera Term삵ĂPC X T[ovOɑ΂āAf[^𑗂B
 * xőȂꍇ́AOobt@Ɋi[AxzB 
 *
 * t[͉L̒ʂB
 * (1) ̃f[^͂B́AOobt@B
 * (2) non-blockingőM݂BSMłreturnB
 * (3) MłȂf[^̓Oobt@֊i[AreturnB
 * (4) ̃f[^͂B
 * (5) Oobt@֊i[AreturnB
 * (6) ̃f[^͂B
 * (7) Oobt@tɂȂAobt@ɎcĂf[^ blocking őM݂B
 *     MsAG[returnB
 * (8) [Uf[^(data/len)̑M blocking őM݂B
 *     MsAG[returnB 
 * (9) MłȂ[Uf[^̓Oobt@֊i[AreturnB
 * (10) Oobt@ɎcĂf[^ non-blockingőM݂B
 *
 * pvar: L\[X
 * buf: Oobt@
 * blocking_write: ^̃pPbgM֐
 * socket: 񓯊^̃\Pbgnh
 * data: f[^
 * len: f[^
 *
 * return TRUE: M
 *        FALSE: MG[
 */
BOOL UTIL_sock_buffered_write(PTInstVar pvar, UTILSockWriteBuf *buf,
                              UTILBlockingWriteCallback blocking_write,
                              SOCKET socket, const char *data,
                              int len)
{
	int curlen;
	int desiredlen;
	int space_required;
	int amount_to_write_from_buffer;
	BOOL did_block = FALSE;
	int first_copy_start;
	int first_copy_amount;

	// ĂяóAȂ炸L if ɓB
	/* Fast path case: buffer is empty, try nonblocking write */
	if (buf->datalen == 0) {
#if 1
		// ܂ non-blocking ŃpPbgMBxł WSAEWOULDBLOCK G[ɂȂA
		// ֐́uMς݃f[^vԂB
		// Ƃ΁AX T[ovOƂ"xterm"NĂꍇAxterm̒[ɉ
		// \ĂԂŁA[̃EBhEhbOƁA֐ 0 ԂĂB
		int sent_amount = send_until_block(pvar, socket, data, len);

		if (sent_amount < 0) {
			return FALSE;
		}
		data += sent_amount;
		len -= sent_amount;
#else
		// mubLO[hőȂꍇAȍ~֑̏AoO邽߁A
		// ܂ƂɓȂB䂦ɁAŃubLO[hgAmɑMĂ܂B
		// |[g](local-to-remote)ɂāAłpPbgMłȂւ̑ΏB
		// (2007.11.29 yutaka)
		// IWiR[hɃoOƎvĂASSHxMɖ肪߁A
		// ꂪ{łAIWiR[hɂ͖肪ȂƍlB
		// {̓mubLOňׂƂAubLOɂ邱ƂɂA
		// Tera TermuȂvAXminguCPUXg[vƂsvcȌۂoĂ܂
		// 悤ɌB̂߁A{̃R[hɖ߂ƂfB
		// (2012.10.14 yutaka)
		if (!blocking_write(pvar, socket, data, len)) {
			return FALSE;
		} else {
			len = 0;
		}
#endif
	}

	// Ăяo non-blocking MŁAׂđ؂ꂽAɐŕԂB
	if (len == 0) {
		return TRUE;
	}

	// Oobt@(buf)ɎcĂf[^ƁAVKMf[^𑫂(desiredlen)A
	// ݂̃obt@(curlen)邩vZB
	//
	// (1)f[^擪Ɋi[ĂP[X
	//
	//                 <----- buflen ------------->
	// buf->bufdata -> +--------------------------+
	//                 |XXXXXXX                   |
	//                 +--------------------------+ 
	//                 <------>
	//                   buf->datalen
	//                 ^
	//                 |
	//                 buf->datastart
	//
	//
	// (2)f[^[Ɋi[ĂP[X
	//
	//                 <----- buflen ------------->
	// buf->bufdata -> +--------------------------+
	//                 |XXXX                  XXXX|
	//                 +--------------------------+ 
	//                 <--->                  <-->
	//                  (a)                    (b)
	//                      (a)+(b) = buf->datalen
	//                                        ^
	//                                        |
	//                                       buf->datastart
	//
	/* We blocked or the buffer has data in it. We need to put this data
	   into the buffer.
	   First, expand buffer as much as possible and necessary. */
	curlen = buf->buflen;
	desiredlen =
		min(pvar->session_settings.WriteBufferSize, buf->datalen + len);

	if (curlen < desiredlen) {
		int wrap_amount = buf->datastart + buf->datalen - curlen;
		int newlen =
			min(pvar->session_settings.WriteBufferSize, 2 * desiredlen);

		buf->bufdata = realloc(buf->bufdata, newlen);
		buf->buflen = newlen;

		if (wrap_amount > 0) {
			int wrap_to_copy = min(wrap_amount, newlen - curlen);

			memmove(buf->bufdata + curlen, buf->bufdata, wrap_to_copy);
			memmove(buf->bufdata, buf->bufdata + wrap_to_copy,
					wrap_amount - wrap_to_copy);
		}
	}

	/* 1) Write data from buffer
	   2) Write data from user
	   3) Copy remaining user data into buffer
	 */
	// obt@ςɂȂAVf[^(data)ꍇ space_required ƂȂB
	// Ȃ킿Aspace_required́uꂽv\B
	//
	//                                               space_required
	//                 <----- buflen -------------><-------->
	// buf->bufdata -> +--------------------------+
	//                 |XXXXXXXXXXXXXXXXX         |
	//                 +--------------------------+ 
	//                 <----------------><------------------>
	//                   buf->datalen            len
	//                 ^
	//                 |
	//                 buf->datastart
	space_required = max(0, buf->datalen + len - buf->buflen);
	amount_to_write_from_buffer = min(buf->datalen, space_required);

	if (amount_to_write_from_buffer > 0) {
		int first_part =
			min(amount_to_write_from_buffer, buf->buflen - buf->datastart);

		did_block = TRUE;
		if (!blocking_write
			(pvar, socket, buf->bufdata + buf->datastart, first_part)) {
			return FALSE;
		}
		if (first_part < amount_to_write_from_buffer) {
			if (!blocking_write
				(pvar, socket, buf->bufdata,
				 amount_to_write_from_buffer - first_part)) {
				return FALSE;
			}
		}

		buf->datalen -= amount_to_write_from_buffer;
		if (buf->datalen == 0) {
			buf->datastart = 0;
		} else {
			buf->datastart =
				(buf->datastart +
				 amount_to_write_from_buffer) % buf->buflen;
		}
		space_required -= amount_to_write_from_buffer;
	}

	if (space_required > 0) {
		did_block = TRUE;
		if (!blocking_write(pvar, socket, data, space_required)) {
			return FALSE;
		}
		data += space_required;
		len -= space_required;
	}

	first_copy_start = (buf->datastart + buf->datalen) % buf->buflen;
	first_copy_amount = min(len, buf->buflen - first_copy_start);
	memcpy(buf->bufdata + first_copy_start, data, first_copy_amount);
	if (first_copy_amount < len) {
		memcpy(buf->bufdata, data + first_copy_amount,
			   len - first_copy_amount);
	}
	buf->datalen += len;

	if (did_block) {
		return UTIL_sock_write_more(pvar, buf, socket);
	} else {
		return TRUE;
	}
}

BOOL UTIL_sock_write_more(PTInstVar pvar, UTILSockWriteBuf *buf,
						  SOCKET socket)
{
	int first_amount = min(buf->buflen - buf->datastart, buf->datalen);
	int sent =
		send_until_block(pvar, socket, buf->bufdata + buf->datastart,
						 first_amount);

	if (sent < 0) {
		return FALSE;
	} else {
		if (sent == first_amount && first_amount < buf->datalen) {
			int sentmore =
				send_until_block(pvar, socket, buf->bufdata,
								 buf->datalen - first_amount);

			if (sentmore < 0) {
				return FALSE;
			}
			sent += sentmore;
		}

		buf->datalen -= sent;
		if (buf->datalen == 0) {
			buf->datastart = 0;
		} else {
			buf->datastart = (buf->datastart + sent) % buf->buflen;
		}
	}

	return TRUE;
}

void UTIL_destroy_sock_write_buf(UTILSockWriteBuf *buf)
{
	SecureZeroMemory(buf->bufdata, buf->buflen);
	buf_destroy(&buf->bufdata, &buf->buflen);
}

BOOL UTIL_is_sock_deeply_buffered(UTILSockWriteBuf *buf)
{
	return buf->buflen / 2 < buf->datalen;
}

void UTIL_get_lang_msg(const char *key, PTInstVar pvar, const char *def)
{
	wchar_t *defW = ToWcharA(def);
	wchar_t *strW;
	GetI18nStrWW("TTSSH", key, defW, pvar->ts->UILanguageFileW, &strW);
	WideCharToACP_t(strW, pvar->UIMsg, sizeof(pvar->UIMsg));
	free(strW);
	free(defW);
}

void UTIL_get_lang_msgU8(const char *key, PTInstVar pvar, const char *def)
{
	const wchar_t *UILanguageFileW = pvar->ts->UILanguageFileW;
	char *strU8;
	GetI18nStrU8W("TTSSH", key, def, UILanguageFileW, &strU8);
	strcpy_s(pvar->UIMsg, MAX_UIMSG, strU8);
	free(strU8);
}

void UTIL_get_lang_msgW(const char *key, PTInstVar pvar, const wchar_t *def, wchar_t *UIMsg)
{
	const wchar_t *UILanguageFileW = pvar->ts->UILanguageFileW;
	wchar_t *strW;
	GetI18nStrWW("TTSSH", key, def, UILanguageFileW, &strW);
	wcscpy_s(UIMsg, MAX_UIMSG, strW);
	free(strW);
}

/*
 *	tHg擾
 *	@retval		tHgnh
 *	@retval		NULL(G[)
 */
HFONT UTIL_get_lang_fixedfont(HWND hWnd, const wchar_t *UILanguageFile)
{
	HFONT hFont;
	LOGFONTW logfont;
	int dpi = GetMonitorDpiFromWindow(hWnd);
	BOOL result = GetI18nLogfontW(L"TTSSH", L"DLG_ABOUT_FONT", &logfont,
								 dpi, UILanguageFile);
	if (result == FALSE) {
		// ǂݍ߂Ȃꍇ͓tHgw肷B
		// GfBbgRg[̓_CAOƓtHgĂ
		// tHgł͂Ȃ߁B

		// EBhE(_CAO)̃tHg擾AtHgQƂ
		HFONT hFontDlg;
		LOGFONTW logfontDlg;
		hFontDlg = (HFONT)SendMessage(hWnd, WM_GETFONT, (WPARAM)0, (LPARAM)0);
		GetObject(hFontDlg, sizeof(logfontDlg), &logfontDlg);

		memset(&logfont, 0, sizeof(logfont));
		wcsncpy_s(logfont.lfFaceName, _countof(logfont.lfFaceName), L"Courier New", _TRUNCATE);
		logfont.lfCharSet = ANSI_CHARSET;	// = 0
		logfont.lfHeight = logfontDlg.lfHeight;
		logfont.lfWidth = 0;
	}
	hFont = CreateFontIndirectW(&logfont);	// G[ NULL
#if 1
	if (hFont == NULL) {
		// tHgłȂꍇ stock object gp
		// DeleteObject() Ăok̂͂
		// It is not necessary (but it is not harmful) to
		// delete stock objects by calling DeleteObject.
		hFont = GetStockObject(ANSI_FIXED_FONT);
	}
#endif
	return hFont;
}
